#ifndef DMRSF_H
#define DMRSF_H

#include "dm.h"
#include "rsf.hpp"
#include "hypercube.h"

namespace Yang {


template<typename T>
class cubeData;

//Abstract class data manager of Regular sampled data, ic oc not determined yet
template <typename T>
class dmRsf: public dm, public rsf<T> {

public:
   dmRsf(const string& itag, dm::typeIO _tIO , const string& szInfo, bool bNoDummy =false): dm(itag, _tIO , szInfo,bNoDummy), rsf<T>() 
   { //if (bInput) //First get the dimension of data from the tag file
	  {
		 //BAD, CALL VIRTUAL FUNCTION in CONSTRUCTOR!!!!
		 //this->getDomain(NULL);
	  }
   }
   
   virtual ~dmRsf() {}
   
   //means init to file history.
   virtual void init(const hypercube* pDom, bool bRemoveDummyAxis = REMOVE_DUMMY_AXIS_DEF) {
      this->getDomain(pDom, bRemoveDummyAxis);
   }
   
   virtual void unInit() {	  
	  //!!! There should be a statement closing the file
	  if(!this->m_bDummy) {
		 auxclose(this->tag.c_str());
	  }
   }

   virtual void seq_grabAll(cubeData<T> * cube) {
	    if(this->dom.sameSize(cube->domain())) {
			//long long n = this->domain().get_n123();
			//Update the counter at first hand
			VERIFY(this->m_seqCounter==0);
		} else {
			seperr("Size incompatible in rsf::seq_grab\n");
		}
   
		 assert(dm::canInput());
		 assert(this->dom.get_ndim()!=0);
		 
		 //update counter in seq_grab function.
		 this->seq_grab(cube);      
		 //reset the counter, the usr should manually do this
		 //this->resetCounter();
   }

   virtual void seq_updateAll(const cubeData<T> * cube){
      if(this->dom.sameSize(cube->domain())) {
        VERIFY(this->m_seqCounter==0);
      }
      else {
        seperr("Size incompatible in rsf::seq_update\n");
      }
      //Load the data here
      this->seq_update(cube);
   }
   
protected:
   //Acquire Domain Size, this function is not virtual
   void getDomain(const hypercube *pDom, bool bRemoveDummyAxis)
   {
    if(NULL == pDom) //Get the domain Information from the tag file
   {
      assert(canInput());
  	  VERIFY(!m_bDummy); //The tag has to be valid.
  	  
  	  //make sure esize matches at least
  	  bool bGrabFail = (0==auxpar("esize","d",&this->esize,this->tag.c_str()));
  	  if(bGrabFail) { //by default, assume it is float
  		  VERIFY(sizeof(T) == sizeof(float));
  		  this->esize = 4;
  	  } else {
  		  VERIFY(sizeof(T) == this->esize);
  	  }
	  
      std::vector<axis> mya;
      int n,i;
	  
      n = getNumDataAxes(this->tag.c_str(), bRemoveDummyAxis);
      
      for(i=1; i <= n; i++)
      {
         mya.push_back(read_axis(i));
      }
      (this->dom).set_axes(mya);
      title=get_string("title","");
    }
    else //class is used for output, set the domain information from the argument
    {
      if (!this->m_bDummy) {
        assert(canOutput());
        if ((this->dom).get_n123()!=0) {
          prt("\n Warn: the dmRsf obj %s(tag=%s) is initialized multiple times", this->m_szInfo.c_str(), this->tag.c_str());
        }
        this->dom = *pDom;
        //once the domain info is decided, write it down right away
        this->writeDomain();
      } else {
        this->dom = *pDom;
      }
    }
    //reset the counter
    this->resetCounter();
    return;
  }
   
  //These following functions deals with SEP-IO
	void writeDomain() //output the domain info to file
	{
	  if(!m_bDummy) {
		  for(int i=1; i <= (this->dom).get_ndim();i++) {
        write_axis(i,(this->dom).get_axis(i));
      }
      //we need to write esize, this is important as well
      int esize = sizeof(T);
      if (this->tag.compare("out")==0) {
         putch("esize","d",&esize);
      } else {
         auxputch("esize","d",&esize,this->tag.c_str());
      }
	  }
	}
	
protected: //all these 4 funcs are part of sequential access function. thus set as private to be used only by the seq_grab_x public interface

   //Read and write data from file to a cubeData object
	void read_all(cubeData<T>& cube)
	{
	long long n1 = (this->dom).get_n123();
	long long n2 = cube.domain().get_n123();
   if(n1 != n2) 
   	seperr("Dataset(%d) and sepcube(%d) not the same size\n", n1, n2);
   
   //here sreed returns only int, means it can maximum support 2GB read/write
   //use sreed_big for large files
   if(int(sizeof(T)*n1)!=sreed(tag.c_str(),cube.pval,n1*sizeof(T))) 
     seperr("Trouble reading data from tag %s\n",tag.c_str());
	}

	void write_all(const cubeData<T>& cube)
	{
	   if(!m_bDummy) {
		   long long n1 = (this->dom).get_n123();
		   long long n2 = cube.domain().get_n123();
		 if(n1!=n2) {
			seperr("Dataset (%d) and sepcube (%d) not the same size\n", n1,n2);
		 }
		 
		 if(int(sizeof(T)*n2) != srite(tag.c_str(),cube.getDataPtr(),n2*sizeof(T))) {
			seperr("Trouble reading data from tag %s\n",tag.c_str());
		 }
	   }
	}
    
protected:
    //In this case, read part of the file into the sep cube data
	void read_next(const cubeData<T>& cube)
	{
      long long n1 = (this->dom).get_n123();
      long long n2 = cube.domain().get_n123();
      if(n1<n2) seperr("Datasize smaller than hypercube\n");
      if(int(sizeof(T)*n2) != sreed(tag.c_str(),cube.pval,n2*sizeof(T))) 
         seperr("Trouble reading data from tag %s\n",tag.c_str());
   }

	void write_next(const cubeData<T>& cube)
	{
	   if(!m_bDummy) {
		   long long n1 = (this->dom).get_n123();
		   long long n2 = cube.domain().get_n123();
		   VERIFY(n2 <= n1);
		 if(int(sizeof(T)*n2)!=srite(tag.c_str(),cube.pval,n2*sizeof(T))) 
			seperr("Trouble reading data from tag %s\n",tag.c_str());
	   }
   }
   
};

//data IO manager for regular sampled data, does not contain array memory in itself, outcore object
template <typename T>
class dmRsf_oc: public dmRsf<T> {

public:
   dmRsf_oc(const string& itag, dm::typeIO _tIO, const string& szInfo, bool bNoDummy=false) : dmRsf<T> (itag,_tIO,szInfo,bNoDummy) {
	  //if (bInput) //First get the dimension of data from the tag file
	  {
		 //BAD, CALL VIRTUAL FUNCTION in CONSTRUCTOR!!!!
		 //this->getDomain(NULL);
	  }
   }
   
   virtual ~dmRsf_oc() { dmRsf<T>::unInit(); }
   	  
   virtual void resetCounter() {
	   rsf<T>::resetCounter();
	   if(!this->m_bDummy) {
		  //Also reset the file pointer
		  int offset=0; int origin=SEEK_SET;
		  VERIFY(0==sseek(this->tag.c_str(),offset,origin));
	   }
	}
   
  //IO access
   
  //Random access, in these cases cube is alloacted by the caller
  virtual void grab3d(cubeData<T> * cube,  const windParam& wind);
  virtual void grab2d(cubeData<T> * cube,  const windParam& wind);
  virtual void grab1d(cubeData<T> * cube,  const windParam& wind);
  virtual void grabXd(cubeData<T> * cube,  const windParam& wind, int ndim);
  //Grab a window of the object
  virtual void grabWindow(cubeData<T> *cube,  const windParam& wind);
  
  virtual void update1d(const cubeData<T> * cube,   const windParam& wind);
  virtual void update2d(const cubeData<T> * cube,   const windParam& wind);
  virtual void update3d(const cubeData<T> * cube,   const windParam& wind);
  virtual void updateXd(const cubeData<T> * cube,   const windParam& wind, int ndim);
  //update a window of the data object 
  virtual void updateWindow(const cubeData<T>* cube,const windParam& wind);
  
  //sequential access
  virtual void seq_grab		(cubeData<T>* cube);
  virtual void seq_update	(const cubeData<T>* cube);
};


//function defintion
template<typename T>
void dmRsf_oc<T>::seq_grab(cubeData<T>* cube) {
   //check size
   rsf<T>::seq_grab(cube);
   this->read_next(*cube);
}
template<typename T>
void dmRsf_oc<T>::seq_update(const cubeData<T>* cube) {
   //check size
   rsf<T>::seq_update(cube);
   this->write_next(*cube);
}
//Random access
template<typename T>
void dmRsf_oc<T>::grab1d(cubeData<T> *cube,  const windParam& wind){
   assert(cube->domain().get_ndim()==1);
   //Literally it is the same using sreed_window function
   this->grabWindow(cube,wind);
}
template<typename T>
void dmRsf_oc<T>::grab2d(cubeData<T> *cube,  const windParam& wind){
   assert(cube->domain().get_ndim()==2);
   //Literally it is the same using sreed_window function
   this->grabWindow(cube,wind);
}
template<typename T>
void dmRsf_oc<T>::grab3d(cubeData<T> *cube,  const windParam& wind){
   //Literally it is the same using sreed_window function
   assert(cube->domain().get_ndim()==3);
   this->grabWindow(cube,wind);
}

template<typename T>
void dmRsf_oc<T>::grabXd(cubeData<T> *cube,  const windParam& wind, int ndim){
   //Literally it is the same using sreed_window function
   assert(cube->domain().get_ndim()==ndim);
   this->grabWindow(cube,wind);
}

template<typename T>
void dmRsf_oc<T>::update1d(const cubeData<T> *cube,  const windParam& wind){
   assert(cube->domain().get_ndim()==1);
   //Literally it is the same using sreed_window function
   this->updateWindow(cube,wind);
}
template<typename T>
void dmRsf_oc<T>::update2d(const cubeData<T> *cube,  const windParam& wind){
   assert(cube->domain().get_ndim()==2);
   //Literally it is the same using sreed_window function
   this->updateWindow(cube,wind);
}
template<typename T>
void dmRsf_oc<T>::update3d(const cubeData<T> *cube,  const windParam& wind){
   assert(cube->domain().get_ndim()==3);
   //Literally it is the same using sreed_window function
   this->updateWindow(cube,wind);
}

template<typename T>
void dmRsf_oc<T>::updateXd(const cubeData<T> *cube,  const windParam& wind, int ndim){
   assert(cube->domain().get_ndim()==ndim);
   //Literally it is the same using sreed_window function
   this->updateWindow(cube,wind);
}

template<typename T>
void dmRsf_oc<T>::grabWindow(cubeData<T> *cube,  const windParam& wind){
   rsf<T>::grabWindow(cube,wind);
   
   int ndim = this->dom.get_ndim();
   stdVecInt ns = this->dom.get_ns();
   stdVecInt nw(ndim,0),fw(ndim,0),jw(ndim,0);   
   //Remember here use the updated windParam
   for (int i=0;i<ndim;i++) {
      nw[i] = this->mWindParamsTmp[i].n;
      fw[i] = this->mWindParamsTmp[i].f;
      jw[i] = this->mWindParamsTmp[i].j;
   }
   sreed_window(this->tag.c_str(),&ndim,&(ns[0]),&(nw[0]),&(fw[0]),&(jw[0]),sizeof(T),(char*)(cube->pval));
}

template<typename T>
void dmRsf_oc<T>::updateWindow(const cubeData<T> *cube,  const windParam& wind){
   rsf<T>::updateWindow(cube,wind);
   
   int ndim = this->dom.get_ndim();
   stdVecInt ns = this->dom.get_ns();
   stdVecInt nw(ndim,0),fw(ndim,0),jw(ndim,0); 
   
   //printWindParam(this->mWindParamsTmp,"Wind in dmRsf_oc update Window");  
   //Remember here use the updated windParam
   for (int i=0;i<ndim;i++) {
      nw[i] = this->mWindParamsTmp[i].n;
      fw[i] = this->mWindParamsTmp[i].f;
      jw[i] = this->mWindParamsTmp[i].j;
   }
   if(!this->m_bDummy) {
	  srite_window(this->tag.c_str(),&ndim,&(ns[0]),&(nw[0]),&(fw[0]),&(jw[0]),sizeof(T),(char*)(cube->pval));
   }
}

//Dummy object just for reading SEP header files but no access to binary
class sepHeaderReader: public dm {
public:
	sepHeaderReader(const string& _tag, typeIO _tIO, const string& szInfo): dm(_tag,_tIO,szInfo,false) 
	{VERIFY(_tIO==dm::IN)}

	void getShape(hypercube &outShape, bool bRemoveDummyAxis=REMOVE_DUMMY_AXIS_DEF)
   {
		//Get the domain Information from the tag file
	  VERIFY(!m_bDummy); //The tag has to be valid.
	  
		std::vector<axis> mya;
      int n,i;
	  
      n = getNumDataAxes(this->tag.c_str(), bRemoveDummyAxis);
      
      for(i=1; i <= n; i++)
      {
         mya.push_back(read_axis(i));
      }
      outShape.set_axes(mya);
      this->title=get_string("title","");
		
		//Get element Size
		bool bGrabFail = (0==auxpar("esize","d",&this->esize,this->tag.c_str()));
		if (bGrabFail) {
			this->esize = 4;
		}
   }
};

} //end of namespace

#endif
