#ifndef IMAGING_H
#define IMAGING_H

#include "cubeData.hpp"

namespace Yang {
  typedef enum _tagMoveOutType {
     constant,
     Linear,
     Quadratic,
     tangentSquare,
     Hyperbolic
   } tagMoveOutType;
  
   typedef tagMoveOutType MOtype;
   
  typedef enum _enumOff2angType {
     seplibOff2ang,
     seplibModified,
     slantstackYang,
     slantstackYangAA //antialias
  } enumOff2angType;
  
  #define DEF_VAL_OFF2ANG2D_N -1
  #define DEF_VAL_OFF2ANG2D_HMIN -1.542e4
  
  void getSlopeAxisFromAngle(axis& ax_p, float amax, float amin, int n);
  
  //Axis are arranged as (z,hx,x) and (z,ax,x)  
  void off2Ang2d(cubeData<float>& out, cubeData<float>& in, bool bBackward, enumOff2angType off2angType, float a_h_max,  
                 int n = DEF_VAL_OFF2ANG2D_N, float a_h_min = DEF_VAL_OFF2ANG2D_HMIN);
  
  //Off <-> angle domain, imaging transform, using slant while avoid fourier transform
  void off2slope2dY(cubeData<float>& out, cubeData<float>& in, bool bAdj);  

  //amp is the model and p1dOut is the data.
  //Place a positive then negative doublet pair on p1dOut based on the coordinates 1 and 2
  //The adjoint would be interpolate p1dOut and subtract then put the result on amp
  inline void placeDoubletLin(float& amp, float* p1dOut, const axis& ax, float coord1, float coord2, bool bAdj){
    //Find the location of coord1 and coord2 on ax.
    float off1 = ax.get_offsetf(coord1);
    float off2 = ax.get_offsetf(coord2);
    //If any of two is out of range, then skip it
    if (off1>0.0f&&off1<float(ax.n-1) && off2>0.0f && off2<float(ax.n-1)) {
      //find wt
      float wt11,wt12,wt21,wt22;
      int ioff1,ioff2;
      ioff1 = int(off1); ioff2 = int(off2);
      wt12 = off1 -ioff1; wt11 = 1.0-wt12;
      wt22 = off2 -ioff2; wt21 = 1.0-wt22;
      if (!bAdj) {
        p1dOut[ioff1]   += amp*wt11;
        p1dOut[ioff1+1] += amp*wt12;
        p1dOut[ioff2]   -= amp*wt21;
        p1dOut[ioff2+1] -= amp*wt22;
      } else {
        amp += wt11*p1dOut[ioff1];
        amp += wt12*p1dOut[ioff1+1];
        amp -= wt21*p1dOut[ioff2];
        amp -= wt22*p1dOut[ioff2+1];
      }
    }
  }

  void off2slope2dYaa(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, float fAntialias, bool bAdd=false, bool bTrueAmpAdj=false);
}

#endif
