#ifndef PARAM_ENTRY_H
#define PARAM_ENTRY_H 1

//Define the string constant for key-value entries in parameter files
typedef const char* const cs;

#ifdef CREATE_PARAM_TEMPLATE
	//define the char* vector to store all the strings
	vector<string> csz_vec;
	vector<string> csz_default_vec;
	
#define DECLARE_CSZ(strVarName,str) \
	cs strVarName=str; \
	csz_vec.push_back(str);\
	csz_default_vec.push_back(string());
	
#define DECLARE_CSZ_DEF(strVarName,str,defVal) \
	cs strVarName=str; \
	csz_vec.push_back(str);\
	csz_default_vec.push_back(defVal);
	
#else
	#define DECLARE_CSZ(strVarName,str) cs strVarName=str;
	
	#define DECLARE_CSZ_DEF(strVarName,str,defVal) \
			DECLARE_CSZ(strVarName,str)	//just ignore the default value
#endif

DECLARE_CSZ(csz_pFreq,"freq") //peak frequency of wavelet

DECLARE_CSZ(csz_totalTime,"T") 
DECLARE_CSZ(csz_dtRec,"dt") 
DECLARE_CSZ(csz_dtSnap,"dt_snap")
DECLARE_CSZ(csz_dtImaging,"dt_imaging")
DECLARE_CSZ(csz_dtComp,"dt_comp")

DECLARE_CSZ(csz_npad,"npad") 
DECLARE_CSZ(csz_damping,"damping")
DECLARE_CSZ(csz_spongeCoef,"spongeCoef")

DECLARE_CSZ(csz_nShots,"nshots") 
DECLARE_CSZ(csz_shot1_ix,"shot1_ix") 
DECLARE_CSZ(csz_shot1_iz,"shot1_iz") 
DECLARE_CSZ(csz_shotIncX,"shot_inc_x") 

//record snapshot params
DECLARE_CSZ(csz_snap_f,"f_snap") 
DECLARE_CSZ(csz_snap_j,"j_snap") 
DECLARE_CSZ(csz_snap_n,"n_snap") 

//obsolete, remove it later
DECLARE_CSZ_DEF(csz_waveletType,"wavelet_type","ricker2") 

DECLARE_CSZ(csz_trace1_ix,"trace1_ix") 
DECLARE_CSZ(csz_trace1_iz,"trace1_iz") 
DECLARE_CSZ(csz_traceIncX,"trace_inc_x") 
DECLARE_CSZ(csz_nTraces,"ntraces")
DECLARE_CSZ(csz_bShot_offset,"bShot_offset")
DECLARE_CSZ_DEF(csz_rollAlong,"rollalong","") 
DECLARE_CSZ_DEF(csz_stenWidth,"stencil_width","5") 
DECLARE_CSZ_DEF(csz_freeSurface,"b_free_surface_top","0") 

DECLARE_CSZ_DEF(csz_bdType,"boundary_type","") 
DECLARE_CSZ_DEF(csz_use4thOrderInTime,"use_4th_order_in_time","") 
DECLARE_CSZ_DEF(csz_isSnapshotIncore,"is_snapshot_incore","") 
DECLARE_CSZ_DEF(csz_isSeisRecIncore,"is_seis_rec_incore","")
DECLARE_CSZ_DEF(csz_actionType,"action_type","")
DECLARE_CSZ_DEF(csz_bLinearModeling,"bLinearModeling","")
DECLARE_CSZ_DEF(csz_take2ndTimeDerivRTM,"take2ndTimeDerivRTM","")

DECLARE_CSZ_DEF(csz_tag_wavelet,"wavelet","") 
DECLARE_CSZ_DEF(csz_tag_wavelet_interped,"wavelet_interped","")
DECLARE_CSZ_DEF(csz_tag_vel,"vel","")

DECLARE_CSZ(csz_tag_srcSnapshot,"src_snap")
DECLARE_CSZ(csz_tag_recvSnapshot,"recv_snap")
DECLARE_CSZ(csz_tag_srcSnapshot2,"src_snap2")
DECLARE_CSZ(csz_tag_recvSnapshot2,"recv_snap2")

DECLARE_CSZ(csz_tag_seisRec,"seis_record")
DECLARE_CSZ(csz_tag_dataObs,"data_obs")

DECLARE_CSZ(csz_tag_images,"images")
DECLARE_CSZ(csz_tag_off_images,"off_images")

DECLARE_CSZ(csz_tag_reflModel,"refl_model")

DECLARE_CSZ(csz_tag_final_image,"final_image")
DECLARE_CSZ(csz_tag_final_off_image,"final_off_image")

/// For Moveout beam focusing inversions
DECLARE_CSZ(csz_subGroupSize,"beam_size")
DECLARE_CSZ(csz_corrStartLag,"start_lag")
DECLARE_CSZ(csz_corrEndLag,"end_lag")

DECLARE_CSZ(csz_bSingleEvent,"bSingleEvent")

DECLARE_CSZ(csz_thresMax,"thresMax")
//For reflection offset Image gathers
DECLARE_CSZ(csz_do_offset,"do_offset")
DECLARE_CSZ(csz_img_dhx,"img_dhx")
DECLARE_CSZ(csz_img_nhx,"img_nhx")
DECLARE_CSZ(csz_img_ohx,"img_ohx")
//For angle domain dso
DECLARE_CSZ(csz_bTaperAngGather,"bTaperAngGather")
DECLARE_CSZ(csz_off2angType,"off2angType")
DECLARE_CSZ(csz_amax,"amax")

//For GPU
DECLARE_CSZ(csz_gpuID,"gpuID")
DECLARE_CSZ(csz_gpuDampingCoef,"damping_gpu")
#endif

