program main
  use sep
  use delay_io_mod
  implicit none

  integer                            :: ngx, ngy, nws, nsx, nsy
  real                               :: dgx, dgy, dws, dsx, dsy
  real                               :: ogx, ogy, ows, osx, osy
  complex, dimension(:)    , pointer :: soufnc
  complex, dimension(:,:,:), pointer :: data, plane_sou, plane_rec
  integer                            :: isx, isy, iws, igx, igy, nrx, npsx, ipsx
  integer                            :: index_sx, index_gx_min, index_gx_max, index_time_delay
  real                               :: gx, gy, sx, sy, ws, sx0, rx, rx_min, rx_max, ot, dt
  real                               :: gx_min, gx_max, sx_min, sx_max, sx0_default
  real                               :: phase_delay, drx, orx, weight
  real                               :: opsx, dpsx, psx, psx_min, psx_max
  logical                            :: offset, encode_sou, encode_rec, wntwgt
  character(len=256)                 :: soufnc_, data_, plane_sou_, plane_rec_
  real, parameter                    :: PI = 3.14159265359

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()

  soufnc_    = "soufnc"
  data_      = "data"
  plane_sou_ = "plane_sou"
  plane_rec_ = "plane_rec"

  call auxin(soufnc_)
  call auxin(data_  )
  call auxout(plane_sou_)
  call auxout(plane_rec_)

  call from_aux(data_, "n1", ngx); call from_aux(data_, "d1", dgx); call from_aux(data_, "o1", ogx)
  call from_aux(data_, "n2", ngy); call from_aux(data_, "d2", dgy); call from_aux(data_, "o2", ogy)
  call from_aux(data_, "n3", nws); call from_aux(data_, "d3", dws); call from_aux(data_, "o3", ows)
  call from_aux(data_, "n4", nsx); call from_aux(data_, "d4", dsx); call from_aux(data_, "o4", osx)
  call from_aux(data_, "n5", nsy); call from_aux(data_, "d5", dsy); call from_aux(data_, "o5", osy)

  call from_param(   "npsx",    npsx,      1 )

  call from_param("wntwgt", wntwgt, .true.)

  call from_param("encode_sou", encode_sou, .true.)
  call from_param("encode_rec", encode_rec, .true.)

  if (npsx==1) then
      call from_param("psx", opsx , 0.   )
      dpsx = 1.
  else
      call from_param("psx_min", psx_min, 0.)
      call from_param("psx_max", psx_max, 0.)
      opsx = psx_min
      dpsx = (psx_max-psx_min)/(npsx-1)
  end if
  call from_param("offset", offset, .false.)

  allocate(soufnc(nws))
  allocate(data(ngx,ngy,nws))

  call sep_read(soufnc, soufnc_)

  write(0,*)"ngx, ngy, nws, nsx, nsy", ngx, ngy, nws, nsx, nsy
  write(0,*)"ogx, ogy, ows, osx, osy", ogx, ogy, ows, osx, osy
  write(0,*)"dgx, dgy, dws, dsx, dsy", dgx, dgy, dws, dsx, dsy


  sx_min = osx; sx_max = osx+(nsx-1)*dsx
  gx_min = ogx; gx_max = ogx+(ngx-1)*dgx
  
  rx_min = gx_min; rx_max = gx_max
 
  if (offset) then
     gx_min = gx_min + sx_min
     gx_max = gx_max + sx_max
  end if 

  rx_min = gx_min; rx_max = gx_max

  if (sx_min<gx_min) rx_min = sx_min
  if (sx_max>gx_max) rx_max = sx_max


  orx = rx_min
  drx = dgx;
  nrx = 1.5 + (rx_max-orx)/drx

  allocate(plane_sou(nrx,1,nws))
  allocate(plane_rec(nrx,1,nws))

  write(0,*)"orx,drx,nrx", orx,drx,nrx

  sx0_default = orx

  call from_param(  "sx0",   sx0,  sx0_default)

  !! now only 2-D

  plane_sou = 0.
  plane_rec = 0.

  gx_min = ogx
  gx_max = ogx+(ngx-1)*dgx

  do ipsx=1, npsx
     psx = opsx+(ipsx-1)*dpsx
     write(0,*)"ipsx=",ipsx,psx,npsx
     plane_sou = 0.
     plane_rec = 0.
     do isx=1, nsx
        sx = osx + (isx-1)*dsx
        gx_min = ogx
        gx_max = ogx+(ngx-1)*dgx
        if (offset) then
           gx_min = gx_min + sx
           gx_max = gx_max + sx
        end if
        index_sx = 1.5+(sx-orx)/drx
        index_gx_min = 1.5+(gx_min-orx)/drx
        index_gx_max = 1.5+(gx_max-orx)/drx
        call iodata(data_, data, ngx, ngy, nws, nsx, nsy, isx, 1, "reed")
        do iws=1, nws
           ws = ows+(iws-1)*dws
           if (wntwgt) then
              weight = sqrt(abs(ws))/nsx     !!.. 2-D case, 3-D: weight=abs(ws)
           else
              weight = 1.
           end if 
           phase_delay = ws*psx*(sx-sx0)
           if (encode_sou) then
              plane_sou(index_sx,1,iws) = soufnc(iws)*cmplx(cos(phase_delay), sin(phase_delay))*weight
           else
              plane_sou(index_sx,1,iws) = soufnc(iws)
           end if
           if (encode_rec) then 
              plane_rec(index_gx_min:index_gx_max,1,iws) = plane_rec(index_gx_min:index_gx_max,1,iws)+data(:,1,iws)*cmplx(cos(phase_delay), sin(phase_delay))*weight
           else
              plane_rec(index_gx_min:index_gx_max,1,iws) = plane_rec(index_gx_min:index_gx_max,1,iws)+data(:,1,iws)
           end if
        end do
     end do
     call sep_write(plane_sou, plane_sou_)
     call sep_write(plane_rec, plane_rec_)
  end do
  call auxputch("n1", "i", nrx, plane_sou_)
  call auxputch("d1", "r", drx, plane_sou_)
  call auxputch("o1", "r", orx, plane_sou_)

  call auxputch("n2", "i", 1  , plane_sou_)
  call auxputch("d2", "r", 1. , plane_sou_)
  call auxputch("o2", "r", 0. , plane_sou_)

  call auxputch("n3", "i", nws, plane_sou_)
  call auxputch("d3", "r", dws, plane_sou_)
  call auxputch("o3", "r", ows, plane_sou_)

  call auxputch("n4", "i", npsx, plane_sou_)
  call auxputch("d4", "r", dpsx, plane_sou_)
  call auxputch("o4", "r", opsx, plane_sou_)

  call auxputch("n5", "i",  1 , plane_sou_)
  call auxputch("d5", "r",  1., plane_sou_)
  call auxputch("o5", "r",  0., plane_sou_)

  call auxputch("esize", "i", 8, plane_sou_)


  call auxputch("n1", "i", nrx, plane_rec_)
  call auxputch("d1", "r", drx, plane_rec_)
  call auxputch("o1", "r", orx, plane_rec_)

  call auxputch("n2", "i", 1  , plane_rec_)
  call auxputch("d2", "r", 1. , plane_rec_)
  call auxputch("o2", "r", 0. , plane_rec_)

  call auxputch("n3", "i", nws, plane_rec_)
  call auxputch("d3", "r", dws, plane_rec_)
  call auxputch("o3", "r", ows, plane_rec_)

  call auxputch("n4", "i", npsx, plane_rec_)
  call auxputch("d4", "r", dpsx, plane_rec_)
  call auxputch("o4", "r", opsx, plane_rec_)

  call auxputch("n5", "i",  1 , plane_rec_)
  call auxputch("d5", "r",  1., plane_rec_)
  call auxputch("o5", "r",  0., plane_rec_)

  call auxputch("esize", "i", 8, plane_rec_)



  call sep_end_prog()
  call exit()


end program
