program main
  use sep
  use lbfgs_bound_mod 
  use wemva_struct_mod
  use wemva_kerspl_mod
  use wemva_offdso_mod
  
  implicit none

  type(img_grid)                              :: igrid
  type(vel_grid)                              :: vgrid
  type(wemva_ftag)                            :: ftag_
  type(axis)                                  :: axk, ayk, azk
  double precision, dimension(:), allocatable :: x, l, u, g, wa
  real            , dimension(:), allocatable :: dspl, dslw, dimg, bslw
  integer         , dimension(:), allocatable :: nbd, iwa
  double precision                            :: f, factr, pgtol
  integer                                     :: nmax, mmax, iprint
  character(len=60)                           :: task, csave
  logical                                     :: lsave(4)
  integer                                     :: isave(44)
  double precision                            :: dsave(29)
  character(len=256)                          :: temp_, spl0_

  integer :: stat, iter, niter, ix, iy, iz, ihx, ihy, ihz, n, m, nm, nh
  real    :: hcut, hord
  logical :: lbfgs_step

  temp_ = "temp"
  spl0_ = "spl0"

  call sep_init(SOURCE)
  call getch_add_string("head=/dev/null")
  call getch_add_string("noheader=y")
  call set_no_putch()
  call sep_begin_prog()
  
  call from_param("niter", niter, 5) 

  call from_param("oxk", axk%o, 0.)
  call from_param("dxk", axk%d, 1.)
  call from_param("nxk", axk%n, 1 )

  call from_param("oyk", ayk%o, 0.)
  call from_param("dyk", ayk%d, 1.)
  call from_param("nyk", ayk%n, 1 )

  call from_param("ozk", azk%o, 0.)
  call from_param("dzk", azk%d, 1.)
  call from_param("nzk", azk%n, 1 )

  call from_param("hcut", hcut, 0.)
  call from_param("hord", hord, 1.)

write(0,*)"111111111111111"

  stat = wemva_kerspl_init(vgrid, igrid, ftag_)
  stat = wemva_offdso_init(igrid%ax%n ,igrid%ay%n ,igrid%az%n ,igrid%ahx%n,igrid%ahy%n, &
                           igrid%ahx%o,igrid%ahy%o,igrid%ahx%d,igrid%ahy%d, hcut, hord)
  stat = base_spln3d_init(vgrid%ax,vgrid%ay,vgrid%az,axk,ayk,azk)

  n = axk%n*ayk%n*azk%n
  m = 7

  iprint = 1
  factr  = 1.0d+4
  pgtol  = 1.0d-6
 
  allocate(x(n), l(n), u(n), g(n), nbd(n))
  allocate(iwa(3*n), wa(2*m*n+4*n+12*m*m+12*m))

  allocate(dspl(axk%n*ayk%n*azk%n))
  allocate(dslw(vgrid%ax%n*vgrid%ay%n*vgrid%az%n))
  allocate(bslw(vgrid%ax%n*vgrid%ay%n*vgrid%az%n))
  allocate(dimg(igrid%ax%n*igrid%ay%n*igrid%az%n))

  call auxinout(spl0_)
  call sseek(spl0_, 0, 0)
  call sep_read(dspl, spl0_)

  x   = dble(dspl) 
  l   = dble(1./8000.)
  u   = dble(1./1500.)
  nbd = 0

  iter = 0
  task = 'START'
  lbfgs_step = .true.

  call auxinout(temp_)
  call wemva_histo_imag(igrid, temp_)
  call wemva_fio_imag_init(igrid, temp_)
 

write(0,*)"start iterate..."
  do while (lbfgs_step .and. iter<=niter)

     lbfgs_step = .false.

     call setulb(n,m,x,l,u,nbd,f,g,factr,pgtol,wa,iwa,task,iprint, &
                 csave,lsave,isave,dsave)

     if (task(1:2)=='FG') then

        dspl = real(x)

        stat = base_spln3d_oper(.false., .false., dspl, bslw)                         !!.. bS = B bM
        call sseek(ftag_%bslw_, 0, 0)
        call sep_write(bslw, ftag_%bslw_)

        !!.. compute the current objfnc and gradient
        stat = wemva_bimpix_oper(         .false., ftag_%bslw_, ftag_%bimg_)          !!.. bI
        stat = wemva_offdso_oper(.false., .false., ftag_%bimg_, ftag_%dimg_)          !!.. D bI
        stat = wemva_offdso_oper(.true. , .false., temp_      , ftag_%dimg_)          !!.. D'D bI
        stat = wemva_kerspl_oper(.true. , .false., ftag_%dspl_, temp_      )          !!.. F'D'D bI

        call sseek(ftag_%dspl_, 0, 0)
        call sep_read(dspl, ftag_%dspl_)

        g = dble(dspl)

        call sseek(ftag_%dimg_, 0, 0)
        f = dble(0.)
        do ihz=1, igrid%ahz%n
        do ihy=1, igrid%ahy%n
        do ihx=1, igrid%ahx%n
           call sep_read(dimg, ftag_%dimg_)
           f = f + dot_product(dimg,dimg)
        end do
        end do
        end do

        write(0,*)"f =", f

        lbfgs_step = .true.

     end if

     if (task(1:5)=='NEW_X') then
        lbfgs_step = .true.
        iter = iter + 1
     end if

  end do

  
end program
